#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stddef.h>
#include <string.h>
#include <omp.h>
#include <mpi.h>
#include "area_common.h"

void calculate_segments(area_info_t* params, long long* segments, double* start, int* rank);

int main(int argc, char **argv) {
    area_info_t params;
    int rank;
    int thread_support = MPI_THREAD_FUNNELED; //indicates that the program
    // is multithreaded, but all MPI operations will be called from the
    // main thread
    params.arch=(char*)"hybrid";
    MPI_Init_thread( &argc, &argv, thread_support, &thread_support );

    // bit of a trick, root rank is 0 which evaluates to false
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    threadRank = rank;

    area_init(argc, argv, &params, rank == 0);

    area_begin(&params, rank == 0);

    // Local copies
    long long segments;
    double start;
    calculate_segments(&params, &segments, &start, &rank);

    double area = 0;
    if( segments > 0 )    
    {
#pragma omp parallel firstprivate(params, segments, start)
        {
            long long i;
#pragma omp for reduction(+:area) schedule(static)
            for(i = 0; i < segments; ++i)
            {
                double x = start + ( i * params.delta );
                double y1 = params.op(x);
                double y2 = params.op(x + params.delta);
                area += (y1 + y2) * .5 * params.delta;
            }
        }
    }
    double total_area;
    MPI_Reduce(&area, &total_area, 1, MPI_DOUBLE, MPI_SUM , 0, MPI_COMM_WORLD);
  
    double cputime = getCPUTime(); 
    MPI_Comm comm = MPI_COMM_WORLD;
    int gsize;
    if (rank == 0){
        MPI_Comm_size(comm, &gsize);
        cputimes = (double *)malloc(gsize*sizeof(double));
    }
    MPI_Gather( &cputime , 1, MPI_DOUBLE, cputimes, 1, MPI_DOUBLE, 0, comm );
    area_end(&params, total_area, rank==0);

    MPI_Finalize();
    area_finish(&params, rank == 0);

    return EXIT_SUCCESS;
}

// structure used to tell the root about ourselves
struct proc_info
{
    char name[MPI_MAX_PROCESSOR_NAME + 1];
    int cores;
};

// for qsorting proc_info pointers
int cmp_proc_info(struct proc_info** left, struct proc_info** right)
{
    // This is reverse from usual so that we get descending order
    if((**left).cores < (**right).cores)
        return 1;
    if((**left).cores == (**right).cores)
        return strcmp((**left).name, (**right).name);
    return -1;
}

// calculate_segments( &params, &segments, &start )
//
// The purpose of this function is to determine how to distribute the segments accross the
// available processes.  To do this we have all processes send their name and # of cores
// to the root process.  The name is sent so that we can detect when 2 or more processes are
// on the same host and so share the same cores.  We then sort based on available cores
// and tell each process what range of threads it's responsible for.
//
// In addition if no # of threads was provided by the user we determine the total number
// of cores and broadcast that to all the processes to be used as the number of threads.
void calculate_segments(area_info_t* params, long long* segments, double* start, int* rank)
{
    int size;
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    // Gather information about this host
    struct proc_info pinfo;
    int name_length;
    MPI_Get_processor_name(pinfo.name, &name_length);
    pinfo.name[name_length] = '\0'; //terminate the string with a null character
    pinfo.cores = omp_get_num_procs();

    // Setup MPI Type for proc_info
    int blocklengths[] = { 1, MPI_MAX_PROCESSOR_NAME + 1, 1, 1 };
    MPI_Aint offsets[] = { 0, offsetof(struct proc_info, name),
                       offsetof(struct proc_info, cores), sizeof(struct proc_info) };
    MPI_Datatype types[] = { MPI_LB, MPI_CHAR, MPI_INT, MPI_UB };
    MPI_Datatype mpi_proc_info;
    MPI_Type_create_struct( sizeof(blocklengths)/sizeof(int), blocklengths, offsets,
                            types, &mpi_proc_info);
    MPI_Type_commit(&mpi_proc_info);

    // Gather host info from all processes
    struct proc_info* info_list = NULL;
    if(rank == 0)
        info_list = (struct proc_info*)alloca(sizeof(struct proc_info) * size);
    MPI_Gather(&pinfo, 1, mpi_proc_info, info_list, 1, mpi_proc_info, 0, MPI_COMM_WORLD);
    
    // List of which cores are assigned to each process.
    // Only used on rank 0
    int (*req_threads_list)[2];

    int threads_used = 0; // How many threads we have assigned

    if(rank == 0)
    {

        // Create array of pointers to be sorted
        struct proc_info** by_cores = (struct proc_info**)alloca(sizeof(struct proc_info*) * size);

        // Populate array
        size_t i;
        for(i = 0; i < size; ++i)
        {
            by_cores[i] = info_list + i;
        }
        // Sort array
        qsort(by_cores, size, sizeof(struct proc_info*),
                (int(*)(const void*, const void*))cmp_proc_info);


        // Populate req_threads_list
        req_threads_list = (int(*)[2])alloca(sizeof(int) * 2 * size);

        for(i = 0; i < size; ++i)
        {
            int req_threads = 0; // Allocate 0 threads by default

            // if this isn't a duplicate entry ( >1 processes on the same host )
             if (i == 0 || strcmp(by_cores[i]->name, by_cores[i - 1]->name) != 0) 
             { 
                 int threads_left = params->threads - threads_used;
                 if( params->threads == 0 || threads_left > by_cores[i]->cores )
                 {
                     req_threads = by_cores[i]->cores;
                 }
                 else
                 {
                     req_threads = threads_left;
                 }

             }

            // Find correct rank index by finding our offset into info_list
            req_threads_list[ by_cores[i] - info_list ][0] = threads_used;
            req_threads_list[ by_cores[i] - info_list ][1] = threads_used + req_threads;

            threads_used += req_threads;
        }

    }
    if( params->threads == 0 )
    {
        if(rank == 0)
            params->threads = threads_used;
        MPI_Bcast(&(params->threads), 1, MPI_INT, 0, MPI_COMM_WORLD);
    }

    int req_cores[2]; // range of cores assigned to this processor
    MPI_Scatter(req_threads_list, 2, MPI_INT, req_cores, 2, MPI_INT, 0, MPI_COMM_WORLD);

    long long segment_chunk = params->segments / params->threads;
    long long rem = params->segments % params->threads;

    *segments = segment_chunk * (req_cores[1] - req_cores[0]) + ( req_cores[1] == params->threads ? rem : 0 );
    *start = params->start + (req_cores[0] * segment_chunk * params->delta);

    if(*segments > 0)
        omp_set_num_threads(req_cores[1] - req_cores[0]);

}
